/****************************************************************************
** Meta object code from reading C++ file 'smsampleedit.hh'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.15.18)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <memory>
#include "./smsampleedit.hh"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'smsampleedit.hh' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.15.18. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_SpectMorph__MainWidget_t {
    QByteArrayData data[12];
    char stringdata0[183];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_SpectMorph__MainWidget_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_SpectMorph__MainWidget_t qt_meta_stringdata_SpectMorph__MainWidget = {
    {
QT_MOC_LITERAL(0, 0, 22), // "SpectMorph::MainWidget"
QT_MOC_LITERAL(1, 23, 17), // "on_volume_changed"
QT_MOC_LITERAL(2, 41, 0), // ""
QT_MOC_LITERAL(3, 42, 10), // "new_volume"
QT_MOC_LITERAL(4, 53, 16), // "on_combo_changed"
QT_MOC_LITERAL(5, 70, 15), // "on_zoom_changed"
QT_MOC_LITERAL(6, 86, 22), // "on_edit_marker_changed"
QT_MOC_LITERAL(7, 109, 21), // "on_mouse_time_changed"
QT_MOC_LITERAL(8, 131, 4), // "time"
QT_MOC_LITERAL(9, 136, 14), // "on_next_sample"
QT_MOC_LITERAL(10, 151, 15), // "on_play_clicked"
QT_MOC_LITERAL(11, 167, 15) // "on_save_clicked"

    },
    "SpectMorph::MainWidget\0on_volume_changed\0"
    "\0new_volume\0on_combo_changed\0"
    "on_zoom_changed\0on_edit_marker_changed\0"
    "on_mouse_time_changed\0time\0on_next_sample\0"
    "on_play_clicked\0on_save_clicked"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_SpectMorph__MainWidget[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       8,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    1,   54,    2, 0x0a /* Public */,
       4,    0,   57,    2, 0x0a /* Public */,
       5,    0,   58,    2, 0x0a /* Public */,
       6,    0,   59,    2, 0x0a /* Public */,
       7,    1,   60,    2, 0x0a /* Public */,
       9,    0,   63,    2, 0x0a /* Public */,
      10,    0,   64,    2, 0x0a /* Public */,
      11,    0,   65,    2, 0x0a /* Public */,

 // slots: parameters
    QMetaType::Void, QMetaType::Int,    3,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Int,    8,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,

       0        // eod
};

void SpectMorph::MainWidget::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<MainWidget *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->on_volume_changed((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 1: _t->on_combo_changed(); break;
        case 2: _t->on_zoom_changed(); break;
        case 3: _t->on_edit_marker_changed(); break;
        case 4: _t->on_mouse_time_changed((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 5: _t->on_next_sample(); break;
        case 6: _t->on_play_clicked(); break;
        case 7: _t->on_save_clicked(); break;
        default: ;
        }
    }
}

QT_INIT_METAOBJECT const QMetaObject SpectMorph::MainWidget::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_meta_stringdata_SpectMorph__MainWidget.data,
    qt_meta_data_SpectMorph__MainWidget,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *SpectMorph::MainWidget::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *SpectMorph::MainWidget::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_SpectMorph__MainWidget.stringdata0))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int SpectMorph::MainWidget::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 8)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 8;
    }
    return _id;
}
struct qt_meta_stringdata_SpectMorph__MainWindow_t {
    QByteArrayData data[1];
    char stringdata0[23];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_SpectMorph__MainWindow_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_SpectMorph__MainWindow_t qt_meta_stringdata_SpectMorph__MainWindow = {
    {
QT_MOC_LITERAL(0, 0, 22) // "SpectMorph::MainWindow"

    },
    "SpectMorph::MainWindow"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_SpectMorph__MainWindow[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

void SpectMorph::MainWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    (void)_o;
    (void)_id;
    (void)_c;
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject SpectMorph::MainWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_meta_stringdata_SpectMorph__MainWindow.data,
    qt_meta_data_SpectMorph__MainWindow,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *SpectMorph::MainWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *SpectMorph::MainWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_SpectMorph__MainWindow.stringdata0))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int SpectMorph::MainWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    return _id;
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
