// Licensed GNU LGPL v2.1 or later: http://www.gnu.org/licenses/lgpl-2.1.html

#ifndef SPECTMORPH_XDG_DIR_HH
#define SPECTMORPH_XDG_DIR_HH

#include <string>

namespace SpectMorph
{

std::string xdg_dir_lookup (const std::string& dir_type);

}

#endif
