/****************************************************************************
** Meta object code from reading C++ file 'smnavigator.hh'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.15.18)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <memory>
#include "./smnavigator.hh"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'smnavigator.hh' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.15.18. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_SpectMorph__Navigator_t {
    QByteArrayData data[21];
    char stringdata0[389];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_SpectMorph__Navigator_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_SpectMorph__Navigator_t qt_meta_stringdata_SpectMorph__Navigator = {
    {
QT_MOC_LITERAL(0, 0, 21), // "SpectMorph::Navigator"
QT_MOC_LITERAL(1, 22, 16), // "wav_data_changed"
QT_MOC_LITERAL(2, 39, 0), // ""
QT_MOC_LITERAL(3, 40, 21), // "show_position_changed"
QT_MOC_LITERAL(4, 62, 21), // "show_analysis_changed"
QT_MOC_LITERAL(5, 84, 27), // "show_frequency_grid_changed"
QT_MOC_LITERAL(6, 112, 13), // "title_changed"
QT_MOC_LITERAL(7, 126, 16), // "on_combo_changed"
QT_MOC_LITERAL(8, 143, 14), // "on_view_sample"
QT_MOC_LITERAL(9, 158, 14), // "on_view_player"
QT_MOC_LITERAL(10, 173, 20), // "on_selection_changed"
QT_MOC_LITERAL(11, 194, 17), // "on_view_time_freq"
QT_MOC_LITERAL(12, 212, 18), // "on_view_fft_params"
QT_MOC_LITERAL(13, 231, 24), // "on_show_position_changed"
QT_MOC_LITERAL(14, 256, 24), // "on_show_analysis_changed"
QT_MOC_LITERAL(15, 281, 30), // "on_show_frequency_grid_changed"
QT_MOC_LITERAL(16, 312, 16), // "on_view_spectrum"
QT_MOC_LITERAL(17, 329, 14), // "on_next_sample"
QT_MOC_LITERAL(18, 344, 14), // "on_prev_sample"
QT_MOC_LITERAL(19, 359, 13), // "on_audio_edit"
QT_MOC_LITERAL(20, 373, 15) // "on_save_clicked"

    },
    "SpectMorph::Navigator\0wav_data_changed\0"
    "\0show_position_changed\0show_analysis_changed\0"
    "show_frequency_grid_changed\0title_changed\0"
    "on_combo_changed\0on_view_sample\0"
    "on_view_player\0on_selection_changed\0"
    "on_view_time_freq\0on_view_fft_params\0"
    "on_show_position_changed\0"
    "on_show_analysis_changed\0"
    "on_show_frequency_grid_changed\0"
    "on_view_spectrum\0on_next_sample\0"
    "on_prev_sample\0on_audio_edit\0"
    "on_save_clicked"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_SpectMorph__Navigator[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
      19,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       5,       // signalCount

 // signals: name, argc, parameters, tag, flags
       1,    0,  109,    2, 0x06 /* Public */,
       3,    0,  110,    2, 0x06 /* Public */,
       4,    0,  111,    2, 0x06 /* Public */,
       5,    0,  112,    2, 0x06 /* Public */,
       6,    0,  113,    2, 0x06 /* Public */,

 // slots: name, argc, parameters, tag, flags
       7,    0,  114,    2, 0x0a /* Public */,
       8,    0,  115,    2, 0x0a /* Public */,
       9,    0,  116,    2, 0x0a /* Public */,
      10,    0,  117,    2, 0x0a /* Public */,
      11,    0,  118,    2, 0x0a /* Public */,
      12,    0,  119,    2, 0x0a /* Public */,
      13,    0,  120,    2, 0x0a /* Public */,
      14,    0,  121,    2, 0x0a /* Public */,
      15,    0,  122,    2, 0x0a /* Public */,
      16,    0,  123,    2, 0x0a /* Public */,
      17,    0,  124,    2, 0x0a /* Public */,
      18,    0,  125,    2, 0x0a /* Public */,
      19,    0,  126,    2, 0x0a /* Public */,
      20,    0,  127,    2, 0x0a /* Public */,

 // signals: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,

       0        // eod
};

void SpectMorph::Navigator::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<Navigator *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->wav_data_changed(); break;
        case 1: _t->show_position_changed(); break;
        case 2: _t->show_analysis_changed(); break;
        case 3: _t->show_frequency_grid_changed(); break;
        case 4: _t->title_changed(); break;
        case 5: _t->on_combo_changed(); break;
        case 6: _t->on_view_sample(); break;
        case 7: _t->on_view_player(); break;
        case 8: _t->on_selection_changed(); break;
        case 9: _t->on_view_time_freq(); break;
        case 10: _t->on_view_fft_params(); break;
        case 11: _t->on_show_position_changed(); break;
        case 12: _t->on_show_analysis_changed(); break;
        case 13: _t->on_show_frequency_grid_changed(); break;
        case 14: _t->on_view_spectrum(); break;
        case 15: _t->on_next_sample(); break;
        case 16: _t->on_prev_sample(); break;
        case 17: _t->on_audio_edit(); break;
        case 18: _t->on_save_clicked(); break;
        default: ;
        }
    } else if (_c == QMetaObject::IndexOfMethod) {
        int *result = reinterpret_cast<int *>(_a[0]);
        {
            using _t = void (Navigator::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Navigator::wav_data_changed)) {
                *result = 0;
                return;
            }
        }
        {
            using _t = void (Navigator::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Navigator::show_position_changed)) {
                *result = 1;
                return;
            }
        }
        {
            using _t = void (Navigator::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Navigator::show_analysis_changed)) {
                *result = 2;
                return;
            }
        }
        {
            using _t = void (Navigator::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Navigator::show_frequency_grid_changed)) {
                *result = 3;
                return;
            }
        }
        {
            using _t = void (Navigator::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Navigator::title_changed)) {
                *result = 4;
                return;
            }
        }
    }
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject SpectMorph::Navigator::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_meta_stringdata_SpectMorph__Navigator.data,
    qt_meta_data_SpectMorph__Navigator,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *SpectMorph::Navigator::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *SpectMorph::Navigator::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_SpectMorph__Navigator.stringdata0))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int SpectMorph::Navigator::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 19)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 19;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 19)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 19;
    }
    return _id;
}

// SIGNAL 0
void SpectMorph::Navigator::wav_data_changed()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void SpectMorph::Navigator::show_position_changed()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void SpectMorph::Navigator::show_analysis_changed()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void SpectMorph::Navigator::show_frequency_grid_changed()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void SpectMorph::Navigator::title_changed()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
